#!/usr/bin/env python
# * coding: utf8 *


import arcpy

class EmptyTest():
    '''A class to find empty geometries
    '''
    def __init__(self, workspace, table_name):
        self.report = {'title': 'Empty Test', 'feature_class': table_name, 'issues': []}
        self.workspace = workspace
        self.table_name = table_name


    def sweep(self):
        '''A method to find empty geometries and return a report dictionarty
        '''
        fields = ['OID@', 'SHAPE@']

        with arcpy.EnvManager(workspace=self.workspace):
            with arcpy.da.SearchCursor(self.table_name, fields) as search_cursor:
                for oid, geometry in search_cursor:
                    if geometry is not None:
                        continue

                    self.report['issues'].append(str(oid))

        return self.report


    def try_fix(self):
        '''A method to that attempts to remove records with empty geometries
        '''
        if len(self.report['issues']) == 0:
            return

        #: for point, polylines, or polygons
        fields = ['OID@']
        query = f'OBJECTID IN ({",".join(self.report["issues"])})'

        with arcpy.EnvManager(workspace=self.workspace):
            with arcpy.da.UpdateCursor(self.table_name, fields, query) as update_cursor:
                for oid, in update_cursor:
                    if oid not in self.report:
                        continue

                    update_cursor.deleteRow()


    def clone(self, table_name):
        print(f'cloning to {table_name}')
        return EmptyTest(self.workspace, table_name)



# -*- coding: utf-8 -*-
'''
Created on Wed Jun 5 14:15:04 2019
@author: eneemann
Script to identify empty geometries in a feature class with open source tools
'''

import time
import fiona
# import os
# import geopandas as gpd

# Start timer and print start time in UTC
start_time = time.time()
readable_start = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())
print('The script start time is {}'.format(readable_start))

# Get list of feature classes in a geodatabase
# database = r'C:\E911\Box Elder CO\BoxElder_Spillman_WGS84.gdb'
database = r'C:\E911\StGeorgeDispatch_TEST\Bad_Geometries_TEST.gdb'
fclist = fiona.listlayers(database)

###############
#  Functions  #
###############


def find_empty_geom(db, data):
    with fiona.open(db, layer=data) as lyr:
        print('working on {} feature class ...'.format(lyr.name,))

        if lyr.schema['geometry'] == ('Point' or 'MultiPoint'):
            del_count = 0
            print('{} geometry type is: {}, looping through features ...'.format(lyr.name, lyr.schema['geometry']))
            # Assign x and y min/max values from lyr bounds and use to verify each point
            xmin, ymin, xmax, ymax = lyr.bounds[0], lyr.bounds[1], lyr.bounds[2], lyr.bounds[3]

            for feature in lyr:
                geom = feature['geometry']
                if geom is None:
                    print('Geometry is invalid (null) for ID: {}'.format(feature['id']))
                    del_count += 1
                # Check for points with coordinates outside of lyr bounds, these are invalid
                elif geom['coordinates'][0] < xmin or geom['coordinates'][0] > xmax:
                    print('Geometry is invalid (outside of bounds) for ID: {}'.format(feature['id']))
                    del_count += 1
                elif geom['coordinates'][1] < ymin or geom['coordinates'][1] > ymax:
                    print('Geometry is invalid (outside of bounds) for ID: {}'.format(feature['id']))
                    del_count += 1

        elif lyr.schema['geometry'] in ['LineString', 'MultiLineString']:
            del_count = 0
            print('{} geometry type is: {}, looping through features ...'.format(lyr.name, lyr.schema['geometry']))
            # Loop through line features, find those with empty coordinate tuples
            for feature in lyr:
                geom = feature['geometry']
                if geom is None:
                    print('Geometry is invalid (null) for ID: {}'.format(feature['id']))
                    del_count += 1
                elif len(geom['coordinates'][0]) == 0:
                    print('Geometry is invalid (empty) for ID: {}'.format(feature['id']))
                    del_count += 1

        elif lyr.schema['geometry'] in ['Polygon', 'MultiPolygon']:
            del_count = 0
            print('{} geometry type is: {}, looping through features ...'.format(lyr.name, lyr.schema['geometry']))
            # Loop through polygon features, find those with empty coordinate tuples
            for feature in lyr:
                geom = feature['geometry']
                if geom is None:
                    print('Geometry is invalid (null) for ID: {}'.format(feature['id']))
                    del_count += 1
                elif len(geom['coordinates'][0]) == 0:
                    print('Geometry is invalid (empty) for ID: {}'.format(feature['id']))
                    del_count += 1

    print('number of feature deleted due to empty geometry: {}'.format(del_count))


##########################
#  Call Functions Below  #
##########################

if __name__ == '__main__':
    for fc in sorted(fclist):
        find_empty_geom(database, fc)

print('Script shutting down ...')
# Stop timer and print end time in UTC
readable_end = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())
print('The script end time is {}'.format(readable_end))
print('Time elapsed: {:.2f}s'.format(time.time() - start_time))


#!/usr/bin/env python
# * coding: utf8 *
'''
address_parser.py
A module that parses street addresses into their various parts.
'''
import json
import pprint
from os.path import dirname, join, realpath

import usaddress

TAG_MAPPING = {
    'AddressNumber': 'address_number',
    # 'AddressNumberPrefix': 'address1',
    'AddressNumberSuffix': 'address_number_suffix',
    'StreetNamePreDirectional': 'prefix_direction',
    'StreetName': 'street_name',
    # 'StreetNamePreModifier': 'address1',
    # 'StreetNamePreType': 'address1',
    'StreetNamePostDirectional': 'street_direction',
    # 'StreetNamePostModifier': 'address1',
    'StreetNamePostType': 'street_type',
    # 'CornerOf': 'address1',
    # 'IntersectionSeparator': 'address1',
    # 'LandmarkName': 'address1',
    # 'USPSBoxGroupID': 'address1',
    # 'USPSBoxGroupType': 'address1',
    # 'USPSBoxID': 'address1',
    # 'USPSBoxType': 'address1',
    # 'BuildingName': 'address2',
    'OccupancyType': 'unit_type',
    'OccupancyIdentifier': 'unit_id',
    # 'SubaddressIdentifier': 'address2',
    # 'SubaddressType': 'address2',
    'PlaceName': 'city',
    # 'StateName': 'state',
    'ZipCode': 'zip_code',
    'USPSBoxID': 'po_box'
}
TWO_CHAR_DIRECTIONS = ['NO', 'SO', 'EA', 'WE']
with open(join(dirname(realpath(__file__)), 'street_types.json'), 'r') as file:
    STREET_TYPES = json.loads(file.read())


class Address():
    '''
    Class for parsing address strings
    '''
    address_number = None
    address_number_suffix = None
    prefix_direction = None
    street_name = None
    street_direction = None
    street_type = None
    unit_type = None
    unit_id = None
    city = None
    zip_code = None
    po_box = None

    def __init__(self, address_text):
        parts, parsed_as = usaddress.tag(address_text.replace('.', ''), TAG_MAPPING)
        if parsed_as not in ['Street Address', 'PO Box']:
            raise Exception(f'{address_text} is not recognized as a valid street address, or P.O. Box')

        for part in parts:
            try:
                value = parts[part].upper()
                if part.endswith('direction'):
                    value = normalize_direction(value)

                setattr(self, part, value)
            except AttributeError:
                pass

        if self.po_box is not None:
            return

        #: look for two-character prefix directions which usaddress does not handle
        street_name_parts = self.street_name.split(' ')
        if len(street_name_parts) > 1:
            if street_name_parts[0].upper() in TWO_CHAR_DIRECTIONS and self.prefix_direction is None:
                self.prefix_direction = normalize_direction(street_name_parts[0])
                self.street_name = ' '.join(street_name_parts[1:])
            elif street_name_parts[-1].upper() in TWO_CHAR_DIRECTIONS and self.street_direction is None:
                self.street_direction = normalize_direction(street_name_parts[-1])
                self.street_name = ' '.join(street_name_parts[:-1])

        if self.street_type is not None:
            self.street_type = normalize_street_type(self.street_type)

        if self.unit_id is not None:
            #: add `#` if there is not unit type
            if not self.unit_id.startswith('#') and self.unit_type is None:
                self.unit_id = f'# {self.unit_id}'

            #: strip `#` if there is a unit type
            elif self.unit_id.startswith('#') and self.unit_type is not None:
                self.unit_id = self.unit_id[1:].strip()

    def __repr__(self):
        properties = vars(self)
        properties.update({'normalized': self.normalized})

        return f'Parsed Address:\n{pprint.pformat(properties)}'

    @property
    def normalized(self):
        '''
        getter for normalized address string
        '''
        if self.po_box is not None:
            return f'PO BOX {self.po_box}'

        parts = [
            self.address_number,
            self.address_number_suffix,
            self.prefix_direction,
            self.street_name,
            self.street_type,
            self.street_direction,
            self.unit_type,
            self.unit_id
        ]

        return ' '.join([part for part in parts if part is not None])


def normalize_direction(direction_text):
    '''
    returns the single letter corresponding to the input direction
    '''

    return direction_text[0].upper()

def normalize_street_type(type_text):
    '''
    returns the standard abbreviation for the input street type
    '''

    type_text = type_text.upper()
    for abbreviation, values in STREET_TYPES.items():
        if type_text in values:
            return abbreviation

    raise InvalidStreetTypeError(type_text)


class InvalidStreetTypeError(Exception):
    '''
    exception for when the street type does not have a corresponding value in street_types.json
    '''
    def __init__(self, type_text):
        super().__init__()
        self.message = f'No matching abbreviation found for {type_text}'



#!/usr/bin/env python
# * coding: utf8 *
'''
sweeper

Usage:
  sweeper sweep duplicates --workspace=<workspace> [--verbose --try-fix --save-report=<report_path> --backup-to=<backup_path> --table-name=<table_name>]
  sweeper sweep empties --workspace=<workspace> [--verbose --try-fix --save-report=<report_path> --backup-to=<backup_path> --table-name=<table_name>]
  sweeper sweep invalids --workspace=<workspace> [--verbose --try-fix --save-report=<report_path> --backup-to=<backup_path> --table-name=<table_name>]
  sweeper sweep --workspace=<workspace> [--verbose --try-fix --save-report=<report_path> --backup-to=<backup_path> --table-name=<table_name>]

Arguments:
  workspace - path to workspace eg: `c:\\my.gdb`
  table_name - name of feature class or table eg: `Roads`
  report_path - folder to save report to eg: `c:\\temp`
  backup_path - place to create a temp gdb and import original table

Examples:
  sweeper sweep --workspace=c:\\data\\thing --try-fix --save-report=c:\\temp --backup-to=c:\\temp\\backup.gdb
'''
import os
import sys

from docopt import docopt

from . import backup, report, workspace_info
from .sweepers.duplicates import DuplicateTest
from .sweepers.empties import EmptyTest


def main():
    '''Main entry point for program. Parse arguments and pass to sweeper modules.
    '''
    args = docopt(__doc__, version='1.0.0')

    #: backup input file before quality checks
    if args['--backup-to']:
        backup.backup_data(args['--workspace'], args['--table-name'], args['--backup-to'])

    #: create a list to hold the instantiated objects.
    closet = []

    #: check what quality check to run.
    if args['duplicates']:
        closet.append(DuplicateTest(args['--workspace'], args['--table-name']))
    elif args['invalids']:
        pass
    elif args['empties']:
        closet.append(EmptyTest(args['--workspace'], args['--table-name']))
    else:
        closet.append(DuplicateTest(args['--workspace'], args['--table-name']))
        closet.append(EmptyTest(args['--workspace'], args['--table-name']))

    reports = execute_sweepers(closet, args['--try-fix'])

    report.print_report(reports)

    if args['--save-report']:
        report.save_report(reports, args['--save-report'])


def execute_sweepers(closet, try_fix):
    '''
    orchestrate the sweeper calls.

    closet: array of sweepers.
    try_fix: bool whether to fix or not.
    '''

    feature_class_names = []
    reports = []

    print(f'running {len(closet)} sweepers. try fix: {try_fix}')
    for tool in closet:
        if tool.table_name:
            report = tool.sweep()

            reports.append(report)

            if try_fix:
                tool.try_fix()
                #: run sweeper again to ensure all errors were fixed.
                report = tool.sweep()

            continue

        print('missing table, executing over workspace')
        #: get all feature classes
        #: explode current tool to match the numbner of feature classes
        #: execute sweep on new exploded tools

        #: get feature class names once
        if len(feature_class_names) == 0:
            feature_class_names = workspace_info.get_featureclasses(tool.workspace)

        #: explode sweeper class for each feature class
        for table_name in feature_class_names:
            new_tool = tool.clone(table_name)
            report = new_tool.sweep()

            reports.append(report)

            if try_fix:
                tool.try_fix()

    return reports


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# * coding: utf8 *
import re

from xxhash import xxh64

import arcpy


class DuplicateTest():
    '''A class that finds and removes duplicate geometries or attributes or both
    '''
    def __init__(self, workspace, table_name):
        self.report = {'title': 'Duplicate Test', 'feature_class': table_name, 'issues': []}
        self.workspace = workspace
        self.table_name = table_name


    def sweep(self):
        '''A method that finds duplicate records and returns a report dictionary
        '''
        digests = set([])

        truncate_shape_precision = re.compile(r'(\d+\.\d{2})(\d+)')

        with arcpy.EnvManager(workspace=self.workspace):
            description = arcpy.da.Describe(self.table_name)

            skip_fields = ['guid', description['shapeFieldName']]

            if description['hasGlobalID']:
                skip_fields.append(description['globalIDFieldName'])

            if description['hasOID']:
                skip_fields.append(description['OIDFieldName'])

            fields = [field.name for field in description['fields'] if field.name not in skip_fields]

            fields.append('SHAPE@WKT')
            fields.append('OID@')

            shapefield_index = fields.index('SHAPE@WKT')
            oid_index = fields.index('OID@')

            with arcpy.da.SearchCursor(self.table_name, fields) as search_cursor:
                for row in search_cursor:
                    shape_wkt = row[shapefield_index]
                    object_id = row[oid_index]

                    if shape_wkt is None:
                        continue

                    #: trim some digits to help with hash matching
                    generalized_wkt = truncate_shape_precision.sub(r'\1', shape_wkt)

                    hasher = xxh64(f'{row[:-2]} {generalized_wkt}')
                    digest = hasher.hexdigest()

                    if digest in digests:
                        self.report['issues'].append(str(object_id))

                    digests.add(digest)

        return self.report


    def try_fix(self):
        '''a method that tries to remove the duplicate records
        '''
        if len(self.report['issues']) == 0:
            return

        sql = f'"OBJECTID" IN ({",".join(self.report["issues"])})'
        temp_feature_layer = 'temp_layer'

        with arcpy.EnvManager(workspace=self.workspace):
            try:
                duplicate_features = arcpy.management.MakeFeatureLayer(self.table_name, temp_feature_layer, sql)

                print(f'attempting to delete {len(self.report["issues"])} duplicate records')

                arcpy.management.DeleteFeatures(duplicate_features)
            except Exception as error:
                print(f'unable to delete features {error}')
            finally:
                if arcpy.Exists(temp_feature_layer):
                    arcpy.management.Delete(temp_feature_layer)


    def clone(self, table_name):
        print(f'cloning to {table_name}')
        return DuplicateTest(self.workspace, table_name)


#!/usr/bin/env python
# * coding: utf8 *
'''
workspace_info.py
A module that gets information about the workspace, including a list of feature classes.
'''
import arcpy
import os


#: A function to generate and return a list of workspace feature classes.
def get_featureclasses(workspace_path):
    '''
    workspace_path: full path to the feature workspace.
    '''

    with arcpy.EnvManager(workspace=workspace_path):
        fc_list = []

        datasets = arcpy.ListDatasets(feature_type='feature')
        datasets = [''] + datasets if datasets is not None else []

        #: generate a list of feature classes showing their full path (Example: Directory/Workspace/Dataset/TableName)
        for ds in datasets:
            for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
                fc_list.append(fc)

        return fc_list



#!/usr/bin/env python
# * coding: utf8 *
'''
report.py
A module that contains the templates for the reports and functions to format data into the report
'''

import os
from datetime import datetime

def _generate_report(writer, reports):
    if len(reports) == 0:
        writer('No issues found!')

        return None

    #: loop through each report dict in the list of dicts
    for report in reports:
        writer(f'{report["title"]} Report for {report["feature_class"]}')
        issues_found = len(report['issues'])
        if issues_found == 0:
            writer('No issues found!')
            writer('---')

            continue

        writer(f'{issues_found} issues found')
        writer('  Issues found:')

        for oid in report['issues']:
            writer(f'    ObjectID {oid}')

        writer('\nSelect statement to view issues in ArcGIS:')
        statement = f'OBJECTID IN ({", ".join(report["issues"])})'

        writer(statement)

        writer('---')

def save_report(reports, save_directory):
    '''
    save_directory 
      - folder sweeper-run-{date}
        - feature class name-sweeper name-numberofissues.txt `Counties-Empties-5.txt`   `Counties-Empties-0.txt`
    '''
    report_directory = _create_report_directory(save_directory)

    for report in reports:
        file_name = _get_file_name(report)
        file_path = os.path.join(report_directory, file_name)

        with open(file_path, 'w') as textfile:
            def write_lines(text):
                textfile.writelines(f'{text}\n')

            _generate_report(write_lines, [report])

def print_report(reports):
    _generate_report(print, reports)

def _get_file_name(report):
    title = report['title'].replace(' ', '')
    
    return f'{report["feature_class"]}_{title}_{len(report["issues"])}.txt'

def _create_report_directory(parent_directory):
    now = datetime.now().strftime('%Y%m%d_%H%M')

    report_directory = os.path.join(parent_directory, f'sweeper_run_{now}')
    os.makedirs(report_directory)

    return report_directory



import glob
import io
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    return io.open(join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")).read()


setup(
    name="agrc-sweeper",
    version="1.0.2",
    license="MIT",
    description="CLI tool for making good data",
    long_description="",
    author="AGRC",
    author_email="agrc@utah.gov",
    url="https://github.com/agrc/sweeper",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(i))[0] for i in glob.glob("src/*.py")],
    python_requires=">=3",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],
    keywords=[],
    install_requires=[
        'docopt==0.6.*',
        'fiona==1.8.*', #: this needs to be installed via conda for now
        'xxhash',
        'usaddress'
    ],
    dependency_links=[],
    extras_require={
        'tests': [
            'pytest',
        ],
        'develop': [
            'yapf',
            'pylint'
        ]
    },
    entry_points={"console_scripts": ["sweeper = sweeper.__main__:main"]}
)
